# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import absolute_import, print_function, unicode_literals

from collections import deque
from datetime import datetime, timedelta
import os
import sys
import threading

if sys.version_info[0] >= 3:
    from io import StringIO
else:
    from StringIO import StringIO

import pywws
import pywws.logger
import pywws.storage
import pywws.template


class UploadThread(threading.Thread):
    def __init__(self, parent, context):
        super(UploadThread, self).__init__()
        self.parent = parent
        self.context = context
        self.queue = deque()

    def run(self):
        self.parent.logger.debug('thread started ' + self.name)
        self.old_message = ''
        if self.context.live_logging:
            polling_interval = self.parent.interval.total_seconds() / 20
            polling_interval = min(max(polling_interval, 4.0), 40.0)
        else:
            polling_interval = 4.0
        while not self.context.shutdown.is_set():
            try:
                OK = self.upload_batch()
            except Exception as ex:
                self.log(str(ex))
                OK = False
            if OK:
                pause = polling_interval
            elif self.context.live_logging:
                # upload failed, wait before trying again
                pause = 40.0
            else:
                # upload failed or nothing more to do
                break
            self.context.shutdown.wait(pause)

    def stop(self):
        if self.is_alive():
            self.parent.logger.debug('stopping thread ' + self.name)
            self.queue.append(None)

    def upload_batch(self):
        if not self.queue:
            return True
        OK = True
        count = 0
        with self.parent.session() as session:
            while self.queue and not self.context.shutdown.is_set():
                if self.parent.catchup == 0:
                    # "live only" service, so ignore old records
                    drop = len(self.queue) - 1
                    if self.queue[-1] is None:
                        drop -= 1
                    if drop > 0:
                        for i in range(drop):
                            self.queue.popleft()
                        self.parent.logger.warning(
                            '{:d} record(s) dropped'.format(drop))
                # send upload without taking it off queue
                upload = self.queue[0]
                if upload is None:
                    OK = False
                    break
                timestamp, kwds = upload
                OK, message = self.parent.upload_data(session, **kwds)
                self.log(message)
                if not OK:
                    break
                count += 1
                if timestamp:
                    self.context.status.set(
                        'last update', self.parent.service_name, str(timestamp))
                # finally remove upload from queue
                self.queue.popleft()
        if self.parent.log_count:
            if count > 1:
                self.parent.logger.warning('{:d} records sent'.format(count))
            elif count:
                self.parent.logger.info('1 record sent')
        return OK

    def log(self, message):
        if message == self.old_message:
            self.parent.logger.debug(message)
        else:
            self.parent.logger.error(message)
            self.old_message = message


class BaseToService(object):
    log_count = True

    def __init__(self, context):
        self.context = context
        # check config
        template = context.params.get(self.service_name, 'template')
        if template == 'default':
            context.params.unset(self.service_name, 'template')
        elif template:
            self.logger.critical(
                'obsolete item "template" found in weather.ini '
                'section [{}]'.format(self.service_name))
        # create templater
        if self.template:
            self.templater = pywws.template.Template(context, use_locale=False)
            self.template_file = StringIO(self.template)
        # get time stamp of last uploaded data
        self.last_update = self.context.status.get_datetime(
            'last update', self.service_name)
        # create upload thread
        self.upload_thread = UploadThread(self, context)
        self.stop = self.upload_thread.stop

    def upload(self, catchup=True, live_data=None, test_mode=False):
        OK = True
        count = 0
        for data, live in self.next_data(catchup and not test_mode, live_data):
            if count >= 30 or len(self.upload_thread.queue) >= 60:
                break
            timestamp = data['idx']
            if test_mode:
                timestamp = None
            prepared_data = self.prepare_data(data)
            prepared_data.update(self.fixed_data)
            self.upload_thread.queue.append(
                (timestamp, {'prepared_data': prepared_data, 'live': live}))
            count += 1
        # start upload thread
        if self.upload_thread.queue and not self.upload_thread.is_alive():
            self.upload_thread.start()

    def prepare_data(self, data):
        data_str = self.templater.make_text(self.template_file, data)
        self.template_file.seek(0)
        return eval('{' + data_str + '}')

    def next_data(self, catchup, live_data):
        if not catchup:
            start = self.context.calib_data.before(datetime.max)
        elif self.last_update:
            start = self.last_update + self.interval
        else:
            start = datetime.utcnow() - max(
                timedelta(days=self.catchup), self.interval)
        if live_data:
            stop = live_data['idx'] - self.interval
        else:
            stop = None
        next_update = start or datetime.min
        for data in self.context.calib_data[start:stop]:
            if data['idx'] >= next_update and self.valid_data(data):
                yield data, False
                self.last_update = data['idx']
                next_update = self.last_update + self.interval
        if (live_data and live_data['idx'] >= next_update and
                self.valid_data(live_data)):
            yield live_data, True
            self.last_update = live_data['idx']

    def valid_data(self, data):
        return True


def main(class_, argv=None):
    import argparse
    import inspect
    if argv is None:
        argv = sys.argv
    docstring = inspect.getdoc(sys.modules[class_.__module__]).split('\n\n')
    parser = argparse.ArgumentParser(
        description=docstring[0], epilog=docstring[1])
    if hasattr(class_, 'register'):
        parser.add_argument('-r', '--register', action='store_true',
                            help='register (or update) with service')
    parser.add_argument('-c', '--catchup', action='store_true',
                        help='upload all data since last upload')
    parser.add_argument('-v', '--verbose', action='count',
                        help='increase amount of reassuring messages')
    parser.add_argument('data_dir', help='root directory of the weather data')
    args = parser.parse_args(argv[1:])
    pywws.logger.setup_handler(args.verbose or 0)
    with pywws.storage.pywws_context(args.data_dir) as context:
        uploader = class_(context)
        if 'register' in args and args.register:
            uploader.register()
            context.flush()
            return 0
        uploader.upload(catchup=args.catchup, test_mode=not args.catchup)
        uploader.stop()
    return 0



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Post a message to Twitter.

Before posting to Twitter you need to authorise pywws by running the
module with the ``-r`` option.

If you run pywws on a low power device such as a Raspberry Pi, you may
find it easier to run this authorisation step on another computer, as
long as it has the required dependencies installed. You can use an empty
'data' directory -- a ``weather.ini`` file will be created whose
contents can be copied into your real ``weather.ini`` file using any
text editor.

Make sure no other pywws software is running, then run the module with
the ``-r`` option::

    python -m pywws.totwitter -r data_dir

(Replace ``data_dir`` with your data directory.)

This will open a web browser window (or give you a URL to copy to your
web browser) where you can log in to your Twitter account and authorise
pywws to post. If the login is successful the browser will display a 7
digit number which you then copy to pywws::

    jim@brains:~/Documents/projects/pywws/master$ python -m pywws.totwitter -r ../data/ 
    12:20:08:pywws.logger:pywws version 18.4.2, build 1521 (487c307)
    Please enter the PIN shown in your web browser: 9069882
    Success! Authorisation data has been stored in ../data/weather.ini
    jim@brains:~/Documents/projects/pywws/master$

The ``secret`` and ``key`` values stored in ``weather.ini`` give access
to your Twitter account and should be kept confidential.

.. _Twitter: https://twitter.com/

"""

from __future__ import absolute_import, print_function

__docformat__ = "restructuredtext en"

import codecs
from contextlib import contextmanager
from datetime import timedelta
import logging
import sys

twitter = None
tweepy = None
try:
    import twitter
except ImportError as ex:
    try:
        import tweepy
    except ImportError:
        # raise exception on the preferred library
        raise ex

from pywws.constants import Twitter as pct
import pywws.localisation
import pywws.logger
import pywws.service
import pywws.storage

logger = logging.getLogger(__name__)


class TweepyHandler(object):
    def __init__(self, key, secret, latitude, longitude):
        logger.info('Using tweepy library')
        auth = tweepy.OAuthHandler(pct.consumer_key, pct.consumer_secret)
        auth.set_access_token(key, secret)
        self.api = tweepy.API(auth)
        if latitude is not None and longitude is not None:
            self.kwargs = {'lat' : latitude, 'long' : longitude}
        else:
            self.kwargs = {}

    def post(self, status, media):
        if len(media) > 1:
            logger.error('Tweepy library cannot post multiple media')
        if media:
            self.api.update_with_media(media[0], status[:257], **self.kwargs)
        else:
            self.api.update_status(status[:280], **self.kwargs)


class PythonTwitterHandler(object):
    def __init__(self, key, secret, latitude, longitude, timeout):
        logger.info('Using python-twitter library')
        self.api = twitter.Api(
            consumer_key=pct.consumer_key,
            consumer_secret=pct.consumer_secret,
            access_token_key=key, access_token_secret=secret,
            timeout=timeout)
        if latitude is not None and longitude is not None:
            self.kwargs = {'latitude' : latitude, 'longitude' : longitude,
                           'display_coordinates' : True}
        else:
            self.kwargs = {}
        self.kwargs['verify_status_length'] = False


    def post(self, status, media):
        max_len = 280
        if media:
            max_len -= len(media[:4]) * 23
        status = status.strip()[:max_len]
        args = dict(self.kwargs)
        if media:
            args['media'] = media
        self.api.PostUpdate(status, **args)


class ToTwitter(object):
    catchup = -1
    interval = timedelta(seconds=150)
    logger = logger
    log_count = True
    service_name = 'pywws.totwitter'

    def __init__(self, context):
        self.params = context.params
        # get parameters
        key = context.params.get('twitter', 'key')
        secret = context.params.get('twitter', 'secret')
        if (not key) or (not secret):
            raise RuntimeError('Authentication data not found')
        latitude = context.params.get('twitter', 'latitude')
        longitude = context.params.get('twitter', 'longitude')
        # open API
        if twitter:
            self.api = PythonTwitterHandler(
                key, secret, latitude, longitude, 40)
        else:
            self.api = TweepyHandler(key, secret, latitude, longitude)
        # create upload thread
        self.upload_thread = pywws.service.UploadThread(self, context)
        self.stop = self.upload_thread.stop

    @contextmanager
    def session(self):
        yield None

    def upload_data(self, session, tweet=''):
        media = []
        while tweet.startswith('media'):
            media_item, tweet = tweet.split('\n', 1)
            media_item = media_item.split()[1]
            media.append(media_item)
        try:
            self.api.post(tweet, media)
        except Exception as ex:
            message = str(ex)
            return 'is a duplicate' in message, message
        return True, 'OK'

    def upload(self, tweet):
        if not tweet:
            return
        self.upload_thread.queue.append((None, {'tweet': tweet}))
        # start upload thread
        if not self.upload_thread.is_alive():
            self.upload_thread.start()

    def upload_file(self, file):
        # get default character encoding of template output
        encoding = self.params.get('config', 'template encoding', 'iso-8859-1')
        with codecs.open(file, 'r', encoding=encoding) as tweet_file:
            tweet = tweet_file.read()
        return self.upload(tweet)


def twitter_auth(params):
    import webbrowser
    if sys.version_info[0] >= 3:
        from urllib.parse import parse_qsl
    else:
        from urlparse import parse_qsl
    import oauth2 as oauth

    consumer = oauth.Consumer(pct.consumer_key, pct.consumer_secret)
    client = oauth.Client(consumer)
    # step 1 - obtain a request token
    resp, content = client.request(
        'https://api.twitter.com/oauth/request_token', 'POST')
    if resp['status'] != '200':
        print('Failed to get request token. [%s]' % resp['status'])
        return 1
    if isinstance(content, bytes):
        content = content.decode('utf-8')
    request_token = dict(parse_qsl(content))
    # step 2 - redirect the user
    redirect_url = 'https://api.twitter.com/oauth/authorize?oauth_token=%s' % (
        request_token['oauth_token'])
    if not webbrowser.open(redirect_url, new=2, autoraise=0):
        print('Please use a web browser to open the following URL')
        print(redirect_url)
    if sys.version_info[0] >= 3:
        input_ = input
    else:
        input_ = raw_input
    pin = input_('Please enter the PIN shown in your web browser: ')
    pin = pin.strip()
    # step 3 - convert the request token to an access token
    token = oauth.Token(
        request_token['oauth_token'], request_token['oauth_token_secret'])
    token.set_verifier(pin)
    client = oauth.Client(consumer, token)
    resp, content = client.request(
        'https://api.twitter.com/oauth/access_token', 'POST')
    if resp['status'] != '200':
        print('Failed to get access token. [%s]' % resp['status'])
        return 1
    if isinstance(content, bytes):
        content = content.decode('utf-8')
    access_token = dict(parse_qsl(content))
    params.set('twitter', 'key', access_token['oauth_token'])
    params.set('twitter', 'secret', access_token['oauth_token_secret'])
    print('Success! Authorisation data has been stored in %s' % params._path)
    return 0


def main(argv=None):
    import argparse
    import inspect
    if argv is None:
        argv = sys.argv
    docstring = inspect.getdoc(sys.modules[__name__]).split('\n\n')
    parser = argparse.ArgumentParser(
        description=docstring[0], epilog=docstring[1])
    parser.add_argument('-r', '--register', action='store_true',
                        help='authorise pywws to post to your account')
    parser.add_argument('-v', '--verbose', action='count',
                        help='increase amount of reassuring messages')
    parser.add_argument('data_dir', help='root directory of the weather data')
    parser.add_argument('file', nargs='*', help='file to be uploaded')
    args = parser.parse_args(argv[1:])
    pywws.logger.setup_handler(args.verbose or 0)
    with pywws.storage.pywws_context(args.data_dir) as context:
        if args.register:
            twitter_auth(context.params)
            context.flush()
            return 0
        uploader = ToTwitter(context)
        for file in args.file:
            uploader.upload_file(file)
        uploader.stop()
    return 0


if __name__ == "__main__":
    sys.exit(main())



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload files to a web server by ftp or copy them to a local directory

This module uploads files to (typically) a website *via* ftp/sftp or
copies files to a local directory (e.g. if you are running pywws on the
web server). Details of the upload destination are stored in the file
``weather.ini`` in your data directory. The only way to set these
details is to edit the file. Run :py:mod:`pywws.towebsite` once to set
the default values, which you can then change. Here is what you're
likely to find when you edit ``weather.ini``::

  [ftp]
  secure = False
  directory = public_html/weather/data/
  local site = False
  password = secret
  site = ftp.username.your_isp.co.uk
  user = username

These are, I hope, fairly obvious. The ``local site`` option lets you
switch from uploading to a remote site to copying to a local site. If
you set ``local site = True`` then you can delete the ``secure``,
``site``, ``user`` and ``password`` lines.

``directory`` is the name of a directory in which all the uploaded
files will be put. This will depend on the structure of your web site
and the sort of host you use. Your hosting provider should be able to
tell you what ``site`` and ``user`` details to use. You should have
already chosen a ``password``.

The ``secure`` option lets you switch from normal ftp to sftp (ftp
over ssh). Some hosting providers offer this as a more secure upload
mechanism, so you should probably use it if available.

"""

from __future__ import absolute_import, print_function

from contextlib import contextmanager
from datetime import timedelta
import logging
import os
import shutil
import sys

if sys.version_info[0] >= 3:
    from io import StringIO
else:
    from StringIO import StringIO

import pywws.logger
import pywws.service
import pywws.storage

__docformat__ = "restructuredtext en"
logger = logging.getLogger(__name__)


class _ftp(object):
    def __init__(self, site, user, password, directory, port):
        global ftplib
        import ftplib
        self.site = site
        self.user = user
        self.password = password
        self.directory = directory
        self.port = port

    @contextmanager
    def session(self):
        logger.info("Uploading to web site with FTP")
        self.ftp = ftplib.FTP()
        self.ftp.connect(self.site, self.port)
        logger.debug('welcome message\n' + self.ftp.getwelcome())
        self.ftp.login(self.user, self.password)
        self.ftp.cwd(self.directory)
        try:
            yield None
        finally:
            self.ftp.close()

    def put(self, src, dest):
        text_file = os.path.splitext(src)[1] in ('.txt', '.xml', '.html')
        if text_file and sys.version_info[0] < 3:
            mode = 'r'
        else:
            mode = 'rb'
        with open(src, mode) as f:
            if text_file:
                self.ftp.storlines('STOR %s' % (dest), f)
            else:
                self.ftp.storbinary('STOR %s' % (dest), f)


class _sftp(object):
    def __init__(self, site, user, password, privkey, directory, port):
        global paramiko
        import paramiko
        self.site = site
        self.user = user
        self.password = password
        self.privkey = privkey
        self.directory = directory
        self.port = port

    @contextmanager
    def session(self):
        logger.info("Uploading to web site with SFTP")
        self.transport = paramiko.Transport((self.site, self.port))
        self.transport.start_client()
        if self.privkey:
            self.get_private_key(self.privkey)
            self.transport.auth_publickey(username=self.user, key=self.pkey)
        else:
            self.transport.auth_password(
                username=self.user, password=self.password)
        self.ftp = paramiko.SFTPClient.from_transport(self.transport)
        self.ftp.chdir(self.directory)
        try:
            yield None
        finally:
            self.ftp.close()
            self.transport.close()

    def put(self, src, dest):
        self.ftp.put(src, dest)

    def get_private_key(self, privkey):
        with open(privkey, 'r') as f:
            s = f.read()
        keyfile = StringIO(s)
        self.pkey = paramiko.RSAKey.from_private_key(keyfile)
        

class _copy(object):
    def __init__(self, directory):
        self.directory = directory

    @contextmanager
    def session(self):
        logger.info("Copying to local directory")
        if not os.path.isdir(self.directory):
            raise RuntimeError(
                'Directory "' + self.directory + '" does not exist.')
        yield None

    def put(self, src, dest):
        shutil.copy2(src, os.path.join(self.directory, dest))


class ToWebSite(object):
    catchup = -1
    interval = timedelta(seconds=150)
    logger = logger
    log_count = False
    service_name = 'pywws.towebsite'

    def __init__(self, context):
        params = context.params
        self.old_ex = None
        if eval(params.get('ftp', 'local site', 'False')):
            # copy to local directory
            directory = params.get(
                'ftp', 'directory',
                os.path.expanduser('~/public_html/weather/data/'))
            self.uploader = _copy(directory)
        else:
            # get remote site details
            site = params.get('ftp', 'site', 'ftp.username.your_isp.co.uk')
            user = params.get('ftp', 'user', 'username')
            # don't set a default password, as might use a private ssh key
            password = params.get('ftp', 'password', '')
            directory = params.get(
                'ftp', 'directory', 'public_html/weather/data/')
            if eval(params.get('ftp', 'secure', 'False')):
                port = eval(params.get('ftp', 'port', '22'))
                privkey = params.get('ftp', 'privkey')
                self.uploader = _sftp(
                    site, user, password, privkey, directory, port)
            else:
                port = eval(params.get('ftp', 'port', '21'))
                self.uploader = _ftp(site, user, password, directory, port)
        # create upload thread
        self.upload_thread = pywws.service.UploadThread(self, context)
        self.stop = self.upload_thread.stop

    @contextmanager
    def session(self):
        with self.uploader.session():
            yield None

    def upload_data(self, session, file='', delete=False):
        if os.path.isfile(file):
            target = os.path.basename(file)
            try:
                self.uploader.put(file, target)
            except Exception as ex:
                return False, str(ex)
            if delete:
                os.unlink(file)
        return True, 'OK'

    def upload(self, files, delete=False):
        if not files:
            return
        for file in files:
            self.upload_thread.queue.append(
                (None, {'file': file, 'delete': delete}))
        # start upload thread
        if not self.upload_thread.is_alive():
            self.upload_thread.start()


def main(argv=None):
    import argparse
    import inspect
    if argv is None:
        argv = sys.argv
    docstring = inspect.getdoc(sys.modules[__name__]).split('\n\n')
    parser = argparse.ArgumentParser(
        description=docstring[0], epilog=docstring[1])
    parser.add_argument('-v', '--verbose', action='count',
                        help='increase amount of reassuring messages')
    parser.add_argument('data_dir', help='root directory of the weather data')
    parser.add_argument('file', nargs='+', help='file to be uploaded')
    args = parser.parse_args(argv[1:])
    pywws.logger.setup_handler(args.verbose or 0)
    with pywws.storage.pywws_context(args.data_dir) as context:
        uploader = ToWebSite(context)
        uploader.upload(args.file)
        uploader.stop()
    return 0


if __name__ == "__main__":
    sys.exit(main())
